# coding: utf-8
from __future__ import absolute_import

from collections import defaultdict
from functools import reduce
from itertools import groupby
from operator import or_

from django.db.models.fields.related import RelatedField
from django.db.models.query_utils import Q
from django.dispatch import Signal
from m3.actions import OperationResult
from objectpack.actions import ObjectDeleteAction
from objectpack.models import ModelProxy

from educommon.utils.db import get_related_fields
from educommon.utils.db import get_related_instances

from .mixins import CascadeDeleteMixin
from .ui import CancelConfirmWindow


class DeleteCheck(object):

    u"""Проверяет объекты перед удалением на наличие зависимостей.

    В случае, если в системе есть объекты, зависимые от удаляемого объекта и
    не отмеченные для каскадного удаления через cascade_delete_objects, то
    отменяет удаление и выводит окно с предупреждением и списком зависимых
    объектов.
    Если в системе есть только зависимые объекты, помеченные
    cascade_delete_objects, то выводится сообщение с кнопкой Удалить, для
    удаления и связанных объектов.

    Для добавления объектов, которые *неявно* зависят от удаляемых, необходимо
    подписаться на сигнал ``DeleteCheck.collect_implicit``.
    """

    #: Сигнал отправляемый после сборки зависимых объектов
    #:
    #: :param objects_to_delete: Удаляемые объекты.
    #: :type objects_to_delete: django.db.models.query.QuerySet
    #: :param related_objects: Словарь, сопоставляющий удаляемые
    #:     и зависимые объекты.
    #: :type related_objects: collections.defaultdict
    collect_implicit = Signal(
        providing_args=['objects_to_delete', 'related_objects']
    )

    def _get_message(self, objects):
        """Возвращает сообщение в правильном склонении."""
        return (
            u'объекта' if len(objects) == 1 else u'объектов',
            u'него' if len(objects) == 1 else u'них'
        )

    def _setup_window(self, objects, related_objects, blocked=True):
        """Возвращает окно с сообщением о связанных объектах.

        :param objects: удаляемые объекты
        :param related_objects: связанные объекты
        :param bool blocked: режим показа окна блокирующее/не блокирующее
        :return: CancelConfirmWindow
        """
        if blocked:
            title = (
                u'Удаление {} невозможно, т.к. на {} есть ссылки:'
                .format(*self._get_message(objects))
            )
        else:
            title = (
                u'При удалении {} будут удалены следующие связи:'
                .format(self._get_message(objects)[0])
            )

        win = CancelConfirmWindow()
        win.set_params(dict(
            title=title,
            objects=objects,
            related_objects=related_objects,
            blocked=blocked
        ))
        win.pack_action_url = self.action.get_absolute_url()
        return win

    def _get_block_relations(self, obj, all_relations, skip_func):
        """Возвращает блокирующие удаление связи.

        :param obj: удаляемый объект
        :param all_relations: все связи объекта
        :param skip_func: функция исключения связи
        :rtype: generator
        """
        for related_obj in all_relations:
            for field in related_obj.__class__._meta.get_fields():
                if (
                    field.concrete and
                    isinstance(field, RelatedField) and
                    getattr(related_obj, field.attname) == obj.pk and
                    (not skip_func or not skip_func(field))
                ):
                    yield related_obj

    def _get_objects(self, context):
        model = self.action.parent.model
        if issubclass(model, ModelProxy):
            model = model.model

        object_ids = getattr(context, self.action.parent.id_param_name)

        return model.objects.filter(pk__in=object_ids)

    def _get_related_instances(self, obj, skip_func=None):
        if isinstance(obj, ModelProxy):
            obj = getattr(obj, obj.model.__name__.lower())
        return get_related_instances(obj, collapse=False, skip_func=skip_func)

    def before(self, request, context):
        if not isinstance(self.action, ObjectDeleteAction):
            return

        # Удаляемые объекты
        objects = self._get_objects(context)

        if not objects:
            return

        if hasattr(self.action.parent, 'skip_field'):
            def skip_function(field):
                return (
                    self.action.parent.skip_field(field) or
                    CascadeDeleteMixin.skip_field(field)
                )
        else:
            skip_function = CascadeDeleteMixin.skip_field

        if getattr(context, 'delete_check_confirmed', False):
            for obj in objects:
                for related_obj in self._get_related_instances(obj):
                    related_obj.delete()

        # Объекты, зависящие от удаляемых
        related_objects = defaultdict(list)
        block_relations = set()
        non_block_relations = set()
        for obj in objects:
            all_relations = set(self._get_related_instances(obj))

            # проверка на наличие блокирующих удаление связей
            block_relations = set(
                self._get_block_relations(obj, all_relations, skip_function)
            )
            for related_obj in block_relations:
                key = '.'.join((
                    related_obj._meta.app_label,
                    related_obj._meta.object_name,
                ))
                related_objects[key].append(related_obj)

            if not related_objects:
                # проверка на наличие неблокирующих удаление связей
                non_block_relations = all_relations - block_relations
                for related_obj in non_block_relations:
                    if getattr(related_obj, 'display_related_error', True):
                        key = '.'.join((
                            related_obj._meta.app_label,
                            related_obj._meta.object_name,
                        ))
                        related_objects[key].append(related_obj)

        self.collect_implicit.send(
            self.action.parent.model,
            objects_to_delete=objects,
            related_objects=related_objects
        )

        if not related_objects:
            return

        win = self._setup_window(objects, related_objects, block_relations)
        win.action_context = context

        return OperationResult(
            success=False,
            code=win.get_script(),
        )
